# fedpkg - a Python library for RPM Packagers
#
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

from bodhi.client.bindings import BodhiClient as _BodhiClient
from fedora.client import AuthError


UPDATE_TYPES = ['bugfix', 'security', 'enhancement', 'newpackage']
REQUEST_TYPES = ['testing', 'stable']
SUGGEST_TYPES = ['unspecified', 'reboot', 'logout']


def clear_csrf_and_retry(func):
    """Clear csrf token and retry

    fedpkg uses Bodhi Python binding API list_overrides first before other
    save and extend APIs. That causes a readonly csrf token is received,
    which will be got again when next time to construct request data to
    modify updates. That is not expected and AuthError will be raised.

    So, the solution is to capture the AuthError error, clear the token and
    try to modify update again by requesting another token with user's
    credential.
    """
    def _decorator(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except AuthError:
            self._session.cookies.clear()
            self.csrf_token = None
            return func(self, *args, **kwargs)
    return _decorator


class BodhiClient(_BodhiClient):
    """Customized BodhiClient for fedpkg"""

    @clear_csrf_and_retry
    def save(self, *args, **kwargs):
        return super(BodhiClient, self).save(*args, **kwargs)

    @clear_csrf_and_retry
    def save_override(self, *args, **kwargs):
        return super(BodhiClient, self).save_override(*args, **kwargs)

    @clear_csrf_and_retry
    def extend_override(self, override, expiration_date):
        data = dict(
            nvr=override['nvr'],
            notes=override['notes'],
            expiration_date=expiration_date,
            csrf_token=self.csrf(),
        )
        return self.send_request(
            'overrides/', verb='POST', auth=True, data=data)
